package controllers

import (
	"context"
	"log/slog"
	"net/http"

	jwtmiddleware "github.com/auth0/go-jwt-middleware/v2"
	"github.com/auth0/go-jwt-middleware/v2/validator"
	"github.com/rs/cors"
	"github.com/urfave/negroni"
)

type Method uint8

const (
	GET Method = iota
	POST
	PUT
	DELETE
)

func (m Method) String() string {
	switch m {
	case GET:
		return "GET"
	case POST:
		return "POST"
	case PUT:
		return "PUT"
	case DELETE:
		return "DELETE"
	default:
		return ""
	}
}

func methodGuard(methods []Method, h http.HandlerFunc) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		for _, m := range methods {
			if r.Method == m.String() {
				h(w, r)
				return
			}
		}

		http.Redirect(w, r, "/404", http.StatusSeeOther)
	}
}

// Register registers all the routes and middlewares.
func Register(appLogger slog.Handler, origin string) (http.Handler, error) {
	r := http.NewServeMux()

	// * Routes

	// * Middlewares
	n := negroni.New()

	logger := negroni.NewLogger()
	logger.ALogger = slog.NewLogLogger(appLogger, slog.LevelInfo)

	static := negroni.NewStatic(http.Dir("public"))
	static.Prefix = "/public"

	jwtValidator, err := validator.New(func(ctx context.Context) (interface{}, error) {
		return []byte("secret"), nil
	},
		validator.HS256,
		origin,
		[]string{"${REPO_NAME_LOWER}-gui"},
		validator.WithCustomClaims(func() validator.CustomClaims {
			return &CustomClaims{}
		}),
	)
	if err != nil {
		return nil, err
	}

	n.Use(negroni.NewRecovery())
	n.Use(logger)
	n.Use(cors.New(cors.Options{
		AllowedOrigins: []string{origin},
	}))
	n.UseHandler(jwtmiddleware.New(jwtValidator.ValidateToken).CheckJWT(nil))
	n.Use(static)
	n.UseHandler(r)

	return n, nil
}
